@@ -68,7 +68,7 @@ const (
 func NewVMIReplicaSet(vmiInformer cache.SharedIndexInformer, vmiRSInformer cache.SharedIndexInformer, recorder record.EventRecorder, clientset kubecli.KubevirtClient, burstReplicas uint) *VMIReplicaSet {
 
 	c := &VMIReplicaSet{
-		Queue:         workqueue.NewRateLimitingQueue(workqueue.DefaultControllerRateLimiter()),
+		Queue:         workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "virt-controller-replicaset"),
 		vmiInformer:   vmiInformer,
 		vmiRSInformer: vmiRSInformer,
 		recorder:      recorder,
@@ -223,11 +223,6 @@ func (c *VMIReplicaSet) execute(key string) error {
 			scaleErr = c.cleanFinishedVmis(rs, finishedVmis)
 		}
 	}
-	// If the controller is going to be deleted and the orphan finalizer is the next one, release the VMIs. Don't update the status
-	// TODO: Workaround for https://github.com/kubernetes/kubernetes/issues/56348, remove it once it is fixed
-	if rs.ObjectMeta.DeletionTimestamp != nil && controller.HasFinalizer(rs, metav1.FinalizerOrphanDependents) {
-		return c.orphan(cm, rs, activeVmis)
-	}
 
 	if scaleErr != nil {
 		logger.Reason(err).Error("Scaling the replicaset failed.")
@@ -241,33 +236,6 @@ func (c *VMIReplicaSet) execute(key string) error {
 	return scaleErr
 }
 
-// orphan removes the owner reference of all VMIs which are owned by the controller instance.
-// Workaround for https://github.com/kubernetes/kubernetes/issues/56348 to make no-cascading deletes possible
-// We don't have to remove the finalizer. This part of the gc is not affected by the mentioned bug
-func (c *VMIReplicaSet) orphan(cm *controller.VirtualMachineControllerRefManager, rs *virtv1.VirtualMachineInstanceReplicaSet, vmis []*virtv1.VirtualMachineInstance) error {
-
-	var wg sync.WaitGroup
-	errChan := make(chan error, len(vmis))
-	wg.Add(len(vmis))
-
-	for _, vmi := range vmis {
-		go func(vmi *virtv1.VirtualMachineInstance) {
-			defer wg.Done()
-			err := cm.ReleaseVirtualMachine(vmi)
-			if err != nil {
-				errChan <- err
-			}
-		}(vmi)
-	}
-	wg.Wait()
-	select {
-	case err := <-errChan:
-		return err
-	default:
-	}
-	return nil
-}
-
 func (c *VMIReplicaSet) scale(rs *virtv1.VirtualMachineInstanceReplicaSet, vmis []*virtv1.VirtualMachineInstance) error {
 	log.Log.V(4).Object(rs).Info("Scale")
 	diff := c.calcDiff(rs, vmis)
@@ -296,7 +264,7 @@ func (c *VMIReplicaSet) scale(rs *virtv1.VirtualMachineInstanceReplicaSet, vmis
 		// We have to delete VMIs, use a very simple selection strategy for now
 		// TODO: Possible deletion order: not yet running VMIs < migrating VMIs < other
 		deleteCandidates := vmis[0:diff]
-		c.expectations.ExpectDeletions(rsKey, controller.VirtualMachineKeys(deleteCandidates))
+		c.expectations.ExpectDeletions(rsKey, controller.VirtualMachineInstanceKeys(deleteCandidates))
 		for i := 0; i < diff; i++ {
 			go func(idx int) {
 				defer wg.Done()
@@ -305,7 +273,7 @@ func (c *VMIReplicaSet) scale(rs *virtv1.VirtualMachineInstanceReplicaSet, vmis
 				// Don't log an error if it is already deleted
 				if err != nil {
 					// We can't observe a delete if it was not accepted by the server
-					c.expectations.DeletionObserved(rsKey, controller.VirtualMachineKey(deleteCandidate))
+					c.expectations.DeletionObserved(rsKey, controller.VirtualMachineInstanceKey(deleteCandidate))
 					c.recorder.Eventf(rs, k8score.EventTypeWarning, FailedDeleteVirtualMachineReason, "Error deleting virtual machine instance %s: %v", deleteCandidate.ObjectMeta.Name, err)
 					errChan <- err
 					return
@@ -585,7 +553,7 @@ func (c *VMIReplicaSet) deleteVirtualMachine(obj interface{}) {
 	if err != nil {
 		return
 	}
-	c.expectations.DeletionObserved(rsKey, controller.VirtualMachineKey(vmi))
+	c.expectations.DeletionObserved(rsKey, controller.VirtualMachineInstanceKey(vmi))
 	c.enqueueReplicaSet(rs)
 }
 
@@ -597,7 +565,7 @@ func (c *VMIReplicaSet) deleteReplicaSet(obj interface{}) {
 	c.enqueueReplicaSet(obj)
 }
 
-func (c *VMIReplicaSet) updateReplicaSet(old, curr interface{}) {
+func (c *VMIReplicaSet) updateReplicaSet(_, curr interface{}) {
 	c.enqueueReplicaSet(curr)
 }
 
@@ -828,7 +796,7 @@ func (c *VMIReplicaSet) cleanFinishedVmis(rs *virtv1.VirtualMachineInstanceRepli
 
 	log.Log.V(4).Object(rs).Info("Delete finished VM's")
 	deleteCandidates := vmis[0:diff]
-	c.expectations.ExpectDeletions(rsKey, controller.VirtualMachineKeys(deleteCandidates))
+	c.expectations.ExpectDeletions(rsKey, controller.VirtualMachineInstanceKeys(deleteCandidates))
 	for i := 0; i < diff; i++ {
 		go func(idx int) {
 			defer wg.Done()
@@ -837,7 +805,7 @@ func (c *VMIReplicaSet) cleanFinishedVmis(rs *virtv1.VirtualMachineInstanceRepli
 			// Don't log an error if it is already deleted
 			if err != nil {
 				// We can't observe a delete if it was not accepted by the server
-				c.expectations.DeletionObserved(rsKey, controller.VirtualMachineKey(deleteCandidate))
+				c.expectations.DeletionObserved(rsKey, controller.VirtualMachineInstanceKey(deleteCandidate))
 				c.recorder.Eventf(rs, k8score.EventTypeWarning, FailedDeleteVirtualMachineReason, "Error deleting finished virtual machine %s: %v", deleteCandidate.ObjectMeta.Name, err)
 				errChan <- err
 				return
