@@ -31,20 +31,15 @@ import (
 	storagev1 "k8s.io/api/storage/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
-	"k8s.io/apimachinery/pkg/types"
-	"k8s.io/apimachinery/pkg/util/sets"
-
-	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1alpha1"
 
 	kubevirtv1 "kubevirt.io/client-go/api/v1"
 	snapshotv1 "kubevirt.io/client-go/apis/snapshot/v1alpha1"
 	"kubevirt.io/client-go/log"
+	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
 	"kubevirt.io/kubevirt/pkg/controller"
 )
 
 const (
-	sourceFinalizer = "snapshot.kubevirt.io/snapshot-source-protection"
-
 	vmSnapshotFinalizer = "snapshot.kubevirt.io/vmsnapshot-protection"
 
 	vmSnapshotContentFinalizer = "snapshot.kubevirt.io/vmsnapshotcontent-protection"
@@ -57,24 +52,11 @@ const (
 
 	volumeSnapshotMissingEvent = "VolumeSnapshotMissing"
 
+	vmSnapshotDeadlineExceededError = "snapshot deadline exceeded"
+
 	snapshotRetryInterval = 5 * time.Second
 )
 
-type snapshotSource interface {
-	UID() types.UID
-	Locked() bool
-	Lock() (bool, error)
-	Unlock() (bool, error)
-	Spec() snapshotv1.SourceSpec
-	PersistentVolumeClaims() map[string]string
-}
-
-type vmSnapshotSource struct {
-	vm         *kubevirtv1.VirtualMachine
-	snapshot   *snapshotv1.VirtualMachineSnapshot
-	controller *VMSnapshotController
-}
-
 func vmSnapshotReady(vmSnapshot *snapshotv1.VirtualMachineSnapshot) bool {
 	return vmSnapshot.Status != nil && vmSnapshot.Status.ReadyToUse != nil && *vmSnapshot.Status.ReadyToUse
 }
@@ -90,8 +72,27 @@ func vmSnapshotError(vmSnapshot *snapshotv1.VirtualMachineSnapshot) *snapshotv1.
 	return nil
 }
 
+func vmSnapshotFailed(vmSnapshot *snapshotv1.VirtualMachineSnapshot) bool {
+	return vmSnapshot.Status != nil && vmSnapshot.Status.Phase == snapshotv1.Failed
+}
+
+func vmSnapshotSucceeded(vmSnapshot *snapshotv1.VirtualMachineSnapshot) bool {
+	return vmSnapshot.Status != nil && vmSnapshot.Status.Phase == snapshotv1.Succeeded
+}
+
 func vmSnapshotProgressing(vmSnapshot *snapshotv1.VirtualMachineSnapshot) bool {
-	return vmSnapshotError(vmSnapshot) == nil && !vmSnapshotReady(vmSnapshot)
+	return vmSnapshotError(vmSnapshot) == nil && !vmSnapshotReady(vmSnapshot) &&
+		!vmSnapshotFailed(vmSnapshot) && !vmSnapshotSucceeded(vmSnapshot)
+}
+
+func vmSnapshotDeadlineExceeded(vmSnapshot *snapshotv1.VirtualMachineSnapshot) bool {
+	if vmSnapshotFailed(vmSnapshot) {
+		return true
+	}
+	if vmSnapshot.Status == nil || vmSnapshot.Status.Phase != snapshotv1.InProgress {
+		return false
+	}
+	return timeUntilDeadline(vmSnapshot) < 0
 }
 
 func getVMSnapshotContentName(vmSnapshot *snapshotv1.VirtualMachineSnapshot) string {
@@ -115,57 +116,74 @@ func translateError(e *vsv1beta1.VolumeSnapshotError) *snapshotv1.Error {
 
 func (ctrl *VMSnapshotController) updateVMSnapshot(vmSnapshot *snapshotv1.VirtualMachineSnapshot) (time.Duration, error) {
 	log.Log.V(3).Infof("Updating VirtualMachineSnapshot %s/%s", vmSnapshot.Namespace, vmSnapshot.Name)
-
-	// Make sure status is initialized
-	if vmSnapshot.Status == nil {
-		return 0, ctrl.updateSnapshotStatus(vmSnapshot, nil)
-	}
+	var retry time.Duration
 
 	source, err := ctrl.getSnapshotSource(vmSnapshot)
 	if err != nil {
 		return 0, err
 	}
 
-	// unlock the source if done/error
-	if !vmSnapshotProgressing(vmSnapshot) && source != nil {
-		if updated, err := source.Unlock(); updated || err != nil {
-			return 0, err
+	content, err := ctrl.getContent(vmSnapshot)
+	if err != nil {
+		return 0, err
+	}
+
+	// Make sure status is initialized before doing anything
+	if vmSnapshot.Status != nil {
+		if source != nil {
+			if vmSnapshotProgressing(vmSnapshot) && vmSnapshot.DeletionTimestamp == nil && !vmSnapshotDeadlineExceeded(vmSnapshot) {
+				// attempt to lock source
+				// if fails will attempt again when source is updated
+				if !source.Locked() {
+					locked, err := source.Lock()
+					if err != nil {
+						return 0, err
+					}
+
+					log.Log.V(3).Infof("Attempt to lock source returned: %t", locked)
+
+					retry = snapshotRetryInterval
+				} else {
+					// create content if does not exist
+					if content == nil {
+						if err := ctrl.createContent(vmSnapshot); err != nil {
+							return 0, err
+						}
+					}
+				}
+			} else {
+				if err := source.Unfreeze(); err != nil {
+					return 0, err
+				}
+
+				if _, err := source.Unlock(); err != nil {
+					return 0, err
+				}
+			}
 		}
 	}
 
-	// check deleted
-	if vmSnapshot.DeletionTimestamp != nil {
-		return 0, ctrl.cleanupVMSnapshot(vmSnapshot)
-	}
+	if (vmSnapshot.DeletionTimestamp != nil || vmSnapshotDeadlineExceeded(vmSnapshot)) && content != nil {
+		if controller.HasFinalizer(content, vmSnapshotContentFinalizer) {
+			cpy := content.DeepCopy()
+			controller.RemoveFinalizer(cpy, vmSnapshotContentFinalizer)
 
-	if source != nil && vmSnapshotProgressing(vmSnapshot) {
-		// attempt to lock source
-		// if fails will attempt again when source is updated
-		if !source.Locked() {
-			locked, err := source.Lock()
+			_, err := ctrl.Client.VirtualMachineSnapshotContent(cpy.Namespace).Update(context.Background(), cpy, metav1.UpdateOptions{})
 			if err != nil {
 				return 0, err
 			}
-
-			log.Log.V(3).Infof("Attempt to lock source returned: %t", locked)
-
-			return snapshotRetryInterval, nil
 		}
 
-		// add source finalizer and maybe other stuff
-		updated, err := ctrl.initVMSnapshot(vmSnapshot)
-		if updated || err != nil {
-			return 0, err
-		}
+		if vmSnapshot.Spec.DeletionPolicy == nil ||
+			*vmSnapshot.Spec.DeletionPolicy == snapshotv1.VirtualMachineSnapshotContentDelete {
+			log.Log.V(2).Infof("Deleting vmsnapshotcontent %s/%s", content.Namespace, content.Name)
 
-		content, err := ctrl.getContent(vmSnapshot)
-		if err != nil {
-			return 0, err
-		}
-
-		// create content if does not exist
-		if content == nil {
-			return 0, ctrl.createContent(vmSnapshot)
+			err = ctrl.Client.VirtualMachineSnapshotContent(vmSnapshot.Namespace).Delete(context.Background(), content.Name, metav1.DeleteOptions{})
+			if err != nil && !errors.IsNotFound(err) {
+				return 0, err
+			}
+		} else {
+			log.Log.V(2).Infof("NOT deleting vmsnapshotcontent %s/%s", content.Namespace, content.Name)
 		}
 	}
 
@@ -173,7 +191,11 @@ func (ctrl *VMSnapshotController) updateVMSnapshot(vmSnapshot *snapshotv1.Virtua
 		return 0, err
 	}
 
-	return 0, nil
+	if retry == 0 {
+		return timeUntilDeadline(vmSnapshot), nil
+	}
+
+	return retry, nil
 }
 
 func (ctrl *VMSnapshotController) updateVMSnapshotContent(content *snapshotv1.VirtualMachineSnapshotContent) (time.Duration, error) {
@@ -181,9 +203,18 @@ func (ctrl *VMSnapshotController) updateVMSnapshotContent(content *snapshotv1.Vi
 
 	var volumeSnapshotStatus []snapshotv1.VolumeSnapshotStatus
 	var deletedSnapshots, skippedSnapshots []string
+	var didFreeze bool
+
+	vmSnapshot, err := ctrl.getVMSnapshot(content)
+	if err != nil || vmSnapshot == nil {
+		return 0, err
+	}
+	if vmSnapshotDeadlineExceeded(vmSnapshot) {
+		return 0, nil
+	}
 
 	currentlyReady := vmSnapshotContentReady(content)
-	currentlyError := content.Status != nil && content.Status.Error != nil
+	currentlyError := (content.Status != nil && content.Status.Error != nil) || vmSnapshotError(vmSnapshot) != nil
 
 	for _, volumeBackup := range content.Spec.VolumeBackups {
 		if volumeBackup.VolumeSnapshotName == nil {
@@ -218,6 +249,36 @@ func (ctrl *VMSnapshotController) updateVMSnapshotContent(content *snapshotv1.Vi
 				continue
 			}
 
+			if !didFreeze {
+				source, err := ctrl.getSnapshotSource(vmSnapshot)
+				if err != nil {
+					return 0, err
+				}
+
+				if source == nil {
+					return 0, fmt.Errorf("unable to get snapshot source")
+				}
+
+				frozen, err := source.Frozen()
+				if err != nil {
+					return 0, err
+				}
+
+				if !frozen {
+					err := source.Freeze()
+					if err != nil {
+						return 0, err
+					}
+
+					// assuming that VM is frozen once Freeze() returns
+					// which should be the case
+					// if Freeze() were async, we'd have to return
+					// and only continue when source.Frozen() == true
+				}
+
+				didFreeze = true
+			}
+
 			volumeSnapshot, err = ctrl.createVolumeSnapshot(content, volumeBackup)
 			if err != nil {
 				return 0, err
@@ -260,8 +321,21 @@ func (ctrl *VMSnapshotController) updateVMSnapshotContent(content *snapshotv1.Vi
 		}
 	}
 
-	if ready && (contentCpy.Status.ReadyToUse == nil || !*contentCpy.Status.ReadyToUse) {
+	if ready && contentCpy.Status.CreationTime == nil {
 		contentCpy.Status.CreationTime = currentTime()
+
+		// TODO revisit with deadline
+		// currently only go into error after becoming ready once
+		source, err := ctrl.getSnapshotSource(vmSnapshot)
+		if err != nil {
+			return 0, err
+		}
+
+		if source != nil {
+			if err := source.Unfreeze(); err != nil {
+				return 0, err
+			}
+		}
 	}
 
 	if errorMessage != "" {
@@ -364,71 +438,6 @@ func (ctrl *VMSnapshotController) getSnapshotSource(vmSnapshot *snapshotv1.Virtu
 	return nil, fmt.Errorf("unknown source %+v", vmSnapshot.Spec.Source)
 }
 
-func (ctrl *VMSnapshotController) initVMSnapshot(vmSnapshot *snapshotv1.VirtualMachineSnapshot) (bool, error) {
-	if controller.HasFinalizer(vmSnapshot, vmSnapshotFinalizer) {
-		return false, nil
-	}
-
-	vmSnapshotCpy := vmSnapshot.DeepCopy()
-	controller.AddFinalizer(vmSnapshotCpy, vmSnapshotFinalizer)
-
-	if _, err := ctrl.Client.VirtualMachineSnapshot(vmSnapshot.Namespace).Update(context.Background(), vmSnapshotCpy, metav1.UpdateOptions{}); err != nil {
-		return false, err
-	}
-
-	return true, nil
-}
-
-func (ctrl *VMSnapshotController) cleanupVMSnapshot(vmSnapshot *snapshotv1.VirtualMachineSnapshot) error {
-	// TODO check restore in progress
-
-	if vmSnapshotProgressing(vmSnapshot) {
-		// will put the snapshot in error state
-		return ctrl.updateSnapshotStatus(vmSnapshot, nil)
-	}
-
-	content, err := ctrl.getContent(vmSnapshot)
-	if err != nil {
-		return err
-	}
-
-	if content != nil {
-		if controller.HasFinalizer(content, vmSnapshotContentFinalizer) {
-			cpy := content.DeepCopy()
-			controller.RemoveFinalizer(cpy, vmSnapshotContentFinalizer)
-
-			_, err := ctrl.Client.VirtualMachineSnapshotContent(cpy.Namespace).Update(context.Background(), cpy, metav1.UpdateOptions{})
-			if err != nil {
-				return err
-			}
-		}
-
-		if vmSnapshot.Spec.DeletionPolicy == nil ||
-			*vmSnapshot.Spec.DeletionPolicy == snapshotv1.VirtualMachineSnapshotContentDelete {
-			log.Log.V(2).Infof("Deleting vmsnapshotcontent %s/%s", content.Namespace, content.Name)
-
-			err = ctrl.Client.VirtualMachineSnapshotContent(vmSnapshot.Namespace).Delete(context.Background(), content.Name, metav1.DeleteOptions{})
-			if err != nil && !errors.IsNotFound(err) {
-				return err
-			}
-		} else {
-			log.Log.V(2).Infof("NOT deleting vmsnapshotcontent %s/%s", content.Namespace, content.Name)
-		}
-	}
-
-	if controller.HasFinalizer(vmSnapshot, vmSnapshotFinalizer) {
-		vmSnapshotCpy := vmSnapshot.DeepCopy()
-		controller.RemoveFinalizer(vmSnapshotCpy, vmSnapshotFinalizer)
-
-		_, err := ctrl.Client.VirtualMachineSnapshot(vmSnapshotCpy.Namespace).Update(context.Background(), vmSnapshotCpy, metav1.UpdateOptions{})
-		if err != nil {
-			return err
-		}
-	}
-
-	return nil
-}
-
 func (ctrl *VMSnapshotController) createContent(vmSnapshot *snapshotv1.VirtualMachineSnapshot) error {
 	source, err := ctrl.getSnapshotSource(vmSnapshot)
 	if err != nil {
@@ -436,7 +445,11 @@ func (ctrl *VMSnapshotController) createContent(vmSnapshot *snapshotv1.VirtualMa
 	}
 
 	var volumeBackups []snapshotv1.VolumeBackup
-	for volumeName, pvcName := range source.PersistentVolumeClaims() {
+	pvcs, err := source.PersistentVolumeClaims()
+	if err != nil {
+		return err
+	}
+	for volumeName, pvcName := range pvcs {
 		pvc, err := ctrl.getSnapshotPVC(vmSnapshot.Namespace, pvcName)
 		if err != nil {
 			return err
@@ -460,6 +473,10 @@ func (ctrl *VMSnapshotController) createContent(vmSnapshot *snapshotv1.VirtualMa
 		volumeBackups = append(volumeBackups, vb)
 	}
 
+	sourceSpec, err := source.Spec()
+	if err != nil {
+		return err
+	}
 	content := &snapshotv1.VirtualMachineSnapshotContent{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:       getVMSnapshotContentName(vmSnapshot),
@@ -468,7 +485,7 @@ func (ctrl *VMSnapshotController) createContent(vmSnapshot *snapshotv1.VirtualMa
 		},
 		Spec: snapshotv1.VirtualMachineSnapshotContentSpec{
 			VirtualMachineSnapshotName: &vmSnapshot.Name,
-			Source:                     source.Spec(),
+			Source:                     sourceSpec,
 			VolumeBackups:              volumeBackups,
 		},
 	}
@@ -574,14 +591,11 @@ func (ctrl *VMSnapshotController) updateSnapshotStatus(vmSnapshot *snapshotv1.Vi
 	}
 
 	if vmSnapshotCpy.DeletionTimestamp != nil {
-		// go into error state
-		if vmSnapshotProgressing(vmSnapshotCpy) {
-			reason := "Snapshot cancelled"
-			vmSnapshotCpy.Status.Error = newError(reason)
-			updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionFalse, reason))
-			updateSnapshotCondition(vmSnapshotCpy, newReadyCondition(corev1.ConditionFalse, reason))
-		}
+		controller.RemoveFinalizer(vmSnapshotCpy, vmSnapshotFinalizer)
 	} else {
+		// since no status subresource can update metadata and status
+		controller.AddFinalizer(vmSnapshotCpy, vmSnapshotFinalizer)
+
 		content, err := ctrl.getContent(vmSnapshot)
 		if err != nil {
 			return err
@@ -589,19 +603,6 @@ func (ctrl *VMSnapshotController) updateSnapshotStatus(vmSnapshot *snapshotv1.Vi
 
 		if content != nil && content.Status != nil {
 			// content exists and is initialized
-			if !vmSnapshotReady(vmSnapshotCpy) && vmSnapshotContentReady(content) {
-				onlineSnapshot, err := ctrl.checkOnlineSnapshotting(vmSnapshotCpy)
-				if err != nil {
-					return err
-				}
-
-				if onlineSnapshot {
-					log.Log.V(3).Infof("Updating snapshot indications that the vm was online")
-					vmSnapshotCpy.Status.Indications = append(vmSnapshotCpy.Status.Indications, snapshotv1.VMSnapshotOnlineSnapshotIndication)
-					log.Log.V(3).Infof("Updating snapshot indications that it was taken without guest agent participation")
-					vmSnapshotCpy.Status.Indications = append(vmSnapshotCpy.Status.Indications, snapshotv1.VMSnapshotNoGuestAgentIndication)
-				}
-			}
 			vmSnapshotCpy.Status.VirtualMachineSnapshotContentName = &content.Name
 			vmSnapshotCpy.Status.CreationTime = content.Status.CreationTime
 			vmSnapshotCpy.Status.ReadyToUse = content.Status.ReadyToUse
@@ -609,18 +610,41 @@ func (ctrl *VMSnapshotController) updateSnapshotStatus(vmSnapshot *snapshotv1.Vi
 		}
 	}
 
-	if vmSnapshotProgressing(vmSnapshotCpy) {
-		source, err := ctrl.getSnapshotSource(vmSnapshot)
-		if err != nil {
-			return err
-		}
-
+	if vmSnapshotDeadlineExceeded(vmSnapshotCpy) {
+		vmSnapshotCpy.Status.Phase = snapshotv1.Failed
+		updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionFalse, vmSnapshotDeadlineExceededError))
+		updateSnapshotCondition(vmSnapshotCpy, newFailureCondition(corev1.ConditionTrue, vmSnapshotDeadlineExceededError))
+	} else if vmSnapshotProgressing(vmSnapshotCpy) {
+		vmSnapshotCpy.Status.Phase = snapshotv1.InProgress
 		if source != nil {
 			if source.Locked() {
 				updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionTrue, "Source locked and operation in progress"))
 			} else {
 				updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionFalse, "Source not locked"))
 			}
+
+			online, err := source.Online()
+			if err != nil {
+				return err
+			}
+
+			indications := []snapshotv1.Indication{}
+			if online {
+				indications = append(indications, snapshotv1.VMSnapshotOnlineSnapshotIndication)
+
+				ga, err := source.GuestAgent()
+				if err != nil {
+					return err
+				}
+
+				if ga {
+					indications = append(indications, snapshotv1.VMSnapshotGuestAgentIndication)
+
+				} else {
+					indications = append(indications, snapshotv1.VMSnapshotNoGuestAgentIndication)
+				}
+			}
+			vmSnapshotCpy.Status.Indications = indications
 		} else {
 			updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionFalse, "Source does not exist"))
 		}
@@ -629,9 +653,11 @@ func (ctrl *VMSnapshotController) updateSnapshotStatus(vmSnapshot *snapshotv1.Vi
 		updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionFalse, "In error state"))
 		updateSnapshotCondition(vmSnapshotCpy, newReadyCondition(corev1.ConditionFalse, "Error"))
 	} else if vmSnapshotReady(vmSnapshotCpy) {
+		vmSnapshotCpy.Status.Phase = snapshotv1.Succeeded
 		updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionFalse, "Operation complete"))
 		updateSnapshotCondition(vmSnapshotCpy, newReadyCondition(corev1.ConditionTrue, "Operation complete"))
 	} else {
+		vmSnapshotCpy.Status.Phase = snapshotv1.Unknown
 		updateSnapshotCondition(vmSnapshotCpy, newProgressingCondition(corev1.ConditionUnknown, "Unknown state"))
 		updateSnapshotCondition(vmSnapshotCpy, newReadyCondition(corev1.ConditionUnknown, "Unknown state"))
 	}
@@ -668,14 +694,26 @@ func (ctrl *VMSnapshotController) getVolumeSnapshotStatus(vm *kubevirtv1.Virtual
 			Reason:  fmt.Sprintf("Volume is nil [%s]", volume.Name),
 		}
 	}
+
+	snapshottable := ctrl.isVolumeSnapshottable(volume)
+	if !snapshottable {
+		return kubevirtv1.VolumeSnapshotStatus{
+			Name:    volume.Name,
+			Enabled: false,
+			Reason:  fmt.Sprintf("Snapshot is not supported for this volumeSource type [%s]", volume.Name),
+		}
+	}
+
 	sc, err := ctrl.getVolumeStorageClass(vm.Namespace, volume)
 	if err != nil {
 		return kubevirtv1.VolumeSnapshotStatus{Name: volume.Name, Enabled: false, Reason: err.Error()}
 	}
+
 	snap, err := ctrl.getVolumeSnapshotClass(sc)
 	if err != nil {
 		return kubevirtv1.VolumeSnapshotStatus{Name: volume.Name, Enabled: false, Reason: err.Error()}
 	}
+
 	if snap == "" {
 		return kubevirtv1.VolumeSnapshotStatus{
 			Name:    volume.Name,
@@ -687,6 +725,11 @@ func (ctrl *VMSnapshotController) getVolumeSnapshotStatus(vm *kubevirtv1.Virtual
 	return kubevirtv1.VolumeSnapshotStatus{Name: volume.Name, Enabled: true}
 }
 
+func (ctrl *VMSnapshotController) isVolumeSnapshottable(volume *kubevirtv1.Volume) bool {
+	return volume.VolumeSource.PersistentVolumeClaim != nil ||
+		volume.VolumeSource.DataVolume != nil
+}
+
 func (ctrl *VMSnapshotController) getVolumeStorageClass(namespace string, volume *kubevirtv1.Volume) (string, error) {
 	// TODO Add Ephemeral (add "|| volume.VolumeSource.Ephemeral != nil" to the `if` below)
 	if volume.VolumeSource.PersistentVolumeClaim != nil {
@@ -706,6 +749,7 @@ func (ctrl *VMSnapshotController) getVolumeStorageClass(namespace string, volume
 		}
 		return "", nil
 	}
+
 	if volume.VolumeSource.DataVolume != nil {
 		storageClassName, err := ctrl.getStorageClassNameForDV(namespace, volume.VolumeSource.DataVolume.Name)
 		if err != nil {
@@ -713,7 +757,8 @@ func (ctrl *VMSnapshotController) getVolumeStorageClass(namespace string, volume
 		}
 		return storageClassName, nil
 	}
-	return "", fmt.Errorf("Volume type has no StorageClass defined")
+
+	return "", fmt.Errorf("volume type has no StorageClass defined")
 }
 
 func (ctrl *VMSnapshotController) getStorageClassNameForDV(namespace string, dvName string) (string, error) {
@@ -812,6 +857,39 @@ func (ctrl *VMSnapshotController) getContent(vmSnapshot *snapshotv1.VirtualMachi
 	return obj.(*snapshotv1.VirtualMachineSnapshotContent).DeepCopy(), nil
 }
 
+func (ctrl *VMSnapshotController) getVMSnapshot(vmSnapshotContent *snapshotv1.VirtualMachineSnapshotContent) (*snapshotv1.VirtualMachineSnapshot, error) {
+	vmSnapshotName := vmSnapshotContent.Spec.VirtualMachineSnapshotName
+	if vmSnapshotName == nil {
+		return nil, fmt.Errorf("VirtualMachineSnapshotName is not initialized in vm snapshot content")
+	}
+
+	obj, exists, err := ctrl.VMSnapshotInformer.GetStore().GetByKey(cacheKeyFunc(vmSnapshotContent.Namespace, *vmSnapshotName))
+	if err != nil || !exists {
+		return nil, err
+	}
+
+	return obj.(*snapshotv1.VirtualMachineSnapshot).DeepCopy(), nil
+}
+
+func (ctrl *VMSnapshotController) getVMI(vm *kubevirtv1.VirtualMachine) (*kubevirtv1.VirtualMachineInstance, bool, error) {
+	key, err := controller.KeyFunc(vm)
+	if err != nil {
+		return nil, false, err
+	}
+
+	obj, exists, err := ctrl.VMIInformer.GetStore().GetByKey(key)
+	if err != nil || !exists {
+		return nil, exists, err
+	}
+
+	return obj.(*kubevirtv1.VirtualMachineInstance).DeepCopy(), true, nil
+}
+
+func (ctrl *VMSnapshotController) checkVMIRunning(vm *kubevirtv1.VirtualMachine) (bool, error) {
+	_, exists, err := ctrl.getVMI(vm)
+	return exists, err
+}
+
 func checkVMRunning(vm *kubevirtv1.VirtualMachine) (bool, error) {
 	rs, err := vm.RunStrategy()
 	if err != nil {
@@ -821,146 +899,6 @@ func checkVMRunning(vm *kubevirtv1.VirtualMachine) (bool, error) {
 	return rs != kubevirtv1.RunStrategyHalted, nil
 }
 
-func (ctrl *VMSnapshotController) checkVMIRunning(vm *kubevirtv1.VirtualMachine) (bool, error) {
-	key, err := controller.KeyFunc(vm)
-	if err != nil {
-		return false, err
-	}
-
-	_, exists, err := ctrl.VMIInformer.GetStore().GetByKey(key)
-	return exists, err
-}
-
-func (ctrl *VMSnapshotController) checkOnlineSnapshotting(vmSnapshot *snapshotv1.VirtualMachineSnapshot) (bool, error) {
-	vm, err := ctrl.getVM(vmSnapshot)
-	if err != nil {
-		return false, err
-	}
-
-	vmRunning, err := checkVMRunning(vm)
-	if err != nil {
-		return false, err
-	}
-
-	exists, err := ctrl.checkVMIRunning(vm)
-	if err != nil {
-		return false, err
-	}
-
-	return (vmRunning || exists), nil
-}
-
-func (s *vmSnapshotSource) UID() types.UID {
-	return s.vm.UID
-}
-
-func (s *vmSnapshotSource) Locked() bool {
-	return s.vm.Status.SnapshotInProgress != nil &&
-		*s.vm.Status.SnapshotInProgress == s.snapshot.Name &&
-		controller.HasFinalizer(s.vm, sourceFinalizer)
-}
-
-func (s *vmSnapshotSource) Lock() (bool, error) {
-	if s.Locked() {
-		return true, nil
-	}
-
-	vmRunning, err := checkVMRunning(s.vm)
-	if err != nil {
-		return false, err
-	}
-
-	exists, err := s.controller.checkVMIRunning(s.vm)
-	if err != nil {
-		return false, err
-	}
-
-	if !vmRunning && !exists {
-		pvcNames := s.pvcNames()
-		pods, err := podsUsingPVCs(s.controller.PodInformer, s.vm.Namespace, pvcNames)
-		if err != nil {
-			return false, err
-		}
-
-		if len(pods) > 0 {
-			log.Log.V(3).Infof("Vm is offline but %d pods using PVCs %+v", len(pods), pvcNames)
-			return false, nil
-		}
-	}
-
-	if s.vm.Status.SnapshotInProgress != nil && *s.vm.Status.SnapshotInProgress != s.snapshot.Name {
-		log.Log.V(3).Infof("Snapshot %s in progress", *s.vm.Status.SnapshotInProgress)
-		return false, nil
-	}
-
-	log.Log.Infof("Adding VM snapshot finalizer to %s", s.vm.Name)
-
-	vmCopy := s.vm.DeepCopy()
-
-	if vmCopy.Status.SnapshotInProgress == nil {
-		vmCopy.Status.SnapshotInProgress = &s.snapshot.Name
-		// unfortunately, status updater does not return the updated resource
-		// but the controller is watching VMs so will get notified
-		return false, s.controller.vmStatusUpdater.UpdateStatus(vmCopy)
-	}
-
-	if !controller.HasFinalizer(vmCopy, sourceFinalizer) {
-		controller.AddFinalizer(vmCopy, sourceFinalizer)
-		_, err = s.controller.Client.VirtualMachine(vmCopy.Namespace).Update(vmCopy)
-		if err != nil {
-			return false, err
-		}
-	}
-
-	return true, nil
-}
-
-func (s *vmSnapshotSource) Unlock() (bool, error) {
-	if s.vm.Status.SnapshotInProgress == nil || *s.vm.Status.SnapshotInProgress != s.snapshot.Name {
-		return false, nil
-	}
-
-	var err error
-	vmCopy := s.vm.DeepCopy()
-
-	if controller.HasFinalizer(vmCopy, sourceFinalizer) {
-		controller.RemoveFinalizer(vmCopy, sourceFinalizer)
-		vmCopy, err = s.controller.Client.VirtualMachine(vmCopy.Namespace).Update(vmCopy)
-		if err != nil {
-			return false, err
-		}
-	}
-
-	vmCopy.Status.SnapshotInProgress = nil
-	err = s.controller.vmStatusUpdater.UpdateStatus(vmCopy)
-	if err != nil {
-		return true, err
-	}
-
-	return true, nil
-}
-
-func (s *vmSnapshotSource) Spec() snapshotv1.SourceSpec {
-	vmCpy := s.vm.DeepCopy()
-	vmCpy.Status = kubevirtv1.VirtualMachineStatus{}
-	return snapshotv1.SourceSpec{
-		VirtualMachine: vmCpy,
-	}
-}
-
-func (s *vmSnapshotSource) PersistentVolumeClaims() map[string]string {
-	return getPVCsFromVolumes(s.vm.Spec.Template.Spec.Volumes)
-}
-
-func (s *vmSnapshotSource) pvcNames() sets.String {
-	pvcs := s.PersistentVolumeClaims()
-	ss := sets.NewString()
-	for _, pvc := range pvcs {
-		ss.Insert(pvc)
-	}
-	return ss
-}
-
 func getPVCsFromVolumes(volumes []kubevirtv1.Volume) map[string]string {
 	pvcs := map[string]string{}
 
