@@ -23,6 +23,7 @@ import (
 	"context"
 	"crypto/tls"
 	"encoding/json"
+	"encoding/xml"
 	"errors"
 	"fmt"
 	"net"
@@ -42,6 +43,16 @@ import (
 	aggregatorclient "k8s.io/kube-aggregator/pkg/client/clientset_generated/clientset"
 	netutils "k8s.io/utils/net"
 
+	"kubevirt.io/kubevirt/tests/framework/matcher"
+
+	"kubevirt.io/kubevirt/tests/libreplicaset"
+
+	"kubevirt.io/kubevirt/tests/util"
+
+	"kubevirt.io/kubevirt/pkg/downwardmetrics/vhostmd/api"
+
+	"kubevirt.io/kubevirt/tests/libvmi"
+
 	k8sv1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/fields"
@@ -76,7 +87,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 	)
 	BeforeEach(func() {
 		virtClient, err = kubecli.GetKubevirtClient()
-		tests.PanicOnError(err)
+		util.PanicOnError(err)
 
 		if aggregatorClient == nil {
 			config, err := kubecli.GetConfig()
@@ -88,6 +99,124 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 		}
 	})
 
+	Describe("changes to the kubernetes client", func() {
+
+		BeforeEach(func() {
+			tests.BeforeTestCleanup()
+		})
+
+		scheduledToRunning := func(vmis []v1.VirtualMachineInstance) time.Duration {
+			var duration time.Duration
+			for _, vmi := range vmis {
+				start := metav1.Time{}
+				stop := metav1.Time{}
+				for _, timestamp := range vmi.Status.PhaseTransitionTimestamps {
+					if timestamp.Phase == v1.Scheduled {
+						start = timestamp.PhaseTransitionTimestamp
+					} else if timestamp.Phase == v1.Running {
+						stop = timestamp.PhaseTransitionTimestamp
+					}
+				}
+				duration += stop.Sub(start.Time)
+			}
+			return duration
+		}
+
+		It("on the controller rate limiter should lead to delayed VMI starts", func() {
+			By("first getting the basetime for a replicaset")
+			replicaset := tests.NewRandomReplicaSetFromVMI(libvmi.NewCirros(libvmi.WithResourceMemory("1Mi")), int32(0))
+			replicaset, err = virtClient.ReplicaSet(util.NamespaceTestDefault).Create(replicaset)
+			Expect(err).ToNot(HaveOccurred())
+			start := time.Now()
+			libreplicaset.DoScaleWithScaleSubresource(virtClient, replicaset.Name, 10)
+			fastDuration := time.Now().Sub(start)
+			libreplicaset.DoScaleWithScaleSubresource(virtClient, replicaset.Name, 0)
+
+			By("reducing the throughput on controller")
+			originalKubeVirt := util.GetCurrentKv(virtClient)
+			originalKubeVirt.Spec.Configuration.ControllerConfiguration = &v1.ReloadableComponentConfiguration{
+				RestClient: &v1.RESTClientConfiguration{
+					RateLimiter: &v1.RateLimiter{
+						TokenBucketRateLimiter: &v1.TokenBucketRateLimiter{
+							Burst: 3,
+							QPS:   2,
+						},
+					},
+				},
+			}
+			tests.UpdateKubeVirtConfigValueAndWait(originalKubeVirt.Spec.Configuration)
+			By("starting a replicaset with reduced throughput")
+			start = time.Now()
+			libreplicaset.DoScaleWithScaleSubresource(virtClient, replicaset.Name, 10)
+			slowDuration := time.Now().Sub(start)
+			Expect(slowDuration.Seconds()).To(BeNumerically(">", 2*fastDuration.Seconds()))
+		})
+
+		It("on the virt handler rate limiter should lead to delayed VMI running states", func() {
+			By("first getting the basetime for a replicaset")
+			targetNode := util.GetAllSchedulableNodes(virtClient).Items[0]
+			vmi := libvmi.NewCirros(
+				libvmi.WithResourceMemory("1Mi"),
+				libvmi.WithNodeSelectorFor(&targetNode),
+			)
+			replicaset := tests.NewRandomReplicaSetFromVMI(vmi, 0)
+			replicaset, err = virtClient.ReplicaSet(util.NamespaceTestDefault).Create(replicaset)
+			Expect(err).ToNot(HaveOccurred())
+			libreplicaset.DoScaleWithScaleSubresource(virtClient, replicaset.Name, 10)
+			Eventually(matcher.AllVMIs(replicaset.Namespace), 90*time.Second, 1*time.Second).Should(matcher.BeInPhase(v1.Running))
+			vmis, err := matcher.AllVMIs(replicaset.Namespace)()
+			Expect(err).ToNot(HaveOccurred())
+			fastDuration := scheduledToRunning(vmis)
+
+			libreplicaset.DoScaleWithScaleSubresource(virtClient, replicaset.Name, 0)
+			Eventually(matcher.AllVMIs(replicaset.Namespace), 90*time.Second, 1*time.Second).Should(matcher.BeGone())
+
+			By("reducing the throughput on handler")
+			originalKubeVirt := util.GetCurrentKv(virtClient)
+			originalKubeVirt.Spec.Configuration.HandlerConfiguration = &v1.ReloadableComponentConfiguration{
+				RestClient: &v1.RESTClientConfiguration{
+					RateLimiter: &v1.RateLimiter{
+						TokenBucketRateLimiter: &v1.TokenBucketRateLimiter{
+							Burst: 1,
+							QPS:   1,
+						},
+					},
+				},
+			}
+			tests.UpdateKubeVirtConfigValueAndWait(originalKubeVirt.Spec.Configuration)
+
+			By("starting a replicaset with reduced throughput")
+			libreplicaset.DoScaleWithScaleSubresource(virtClient, replicaset.Name, 10)
+			Eventually(matcher.AllVMIs(replicaset.Namespace), 180*time.Second, 1*time.Second).Should(matcher.BeInPhase(v1.Running))
+			vmis, err = matcher.AllVMIs(replicaset.Namespace)()
+			Expect(err).ToNot(HaveOccurred())
+			slowDuration := scheduledToRunning(vmis)
+			Expect(slowDuration.Seconds()).To(BeNumerically(">", 1.5*fastDuration.Seconds()))
+		})
+	})
+
+	Describe("downwardMetrics", func() {
+		It("[test_id:6535]should be published to a vmi and periodically updated", func() {
+			vmi := libvmi.NewTestToolingFedora()
+			tests.AddDownwardMetricsVolume(vmi, "vhostmd")
+			vmi = tests.RunVMIAndExpectLaunch(vmi, 180)
+			Expect(console.LoginToFedora(vmi)).To(Succeed())
+
+			metrics, err := getDownwardMetrics(vmi)
+			Expect(err).ToNot(HaveOccurred())
+			timestamp := getTimeFromMetrics(metrics)
+
+			vmi, err = virtClient.VirtualMachineInstance(vmi.Namespace).Get(vmi.Name, &metav1.GetOptions{})
+			Expect(err).ToNot(HaveOccurred())
+			Eventually(func() int {
+				metrics, err = getDownwardMetrics(vmi)
+				Expect(err).ToNot(HaveOccurred())
+				return getTimeFromMetrics(metrics)
+			}, 10*time.Second, 1*time.Second).ShouldNot(Equal(timestamp))
+			Expect(getHostnameFromMetrics(metrics)).To(Equal(vmi.Status.NodeName))
+		})
+	})
+
 	Describe("CRDs", func() {
 		It("[test_id:5177]Should have structural schema", func() {
 			ourCRDs := []string{crds.VIRTUALMACHINE, crds.VIRTUALMACHINEINSTANCE, crds.VIRTUALMACHINEINSTANCEPRESET,
@@ -179,11 +308,11 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 
 			By("checking that we can still start virtual machines and connect to the VMI")
 			vmi := tests.NewRandomVMIWithEphemeralDisk(cd.ContainerDiskFor(cd.ContainerDiskAlpine))
-			vmi = tests.RunVMIAndExpectLaunch(vmi, 60)
+			vmi = tests.RunVMI(vmi, 60)
 			Expect(console.LoginToAlpine(vmi)).To(Succeed())
 		})
 
-		It("[QUARANTINE][sig-compute][test_id:4100] should be valid during the whole rotation process", func() {
+		It("[sig-compute][test_id:4100] should be valid during the whole rotation process", func() {
 			oldAPICert := tests.EnsurePodsCertIsSynced(fmt.Sprintf("%s=%s", v1.AppLabel, "virt-api"), flags.KubeVirtInstallNamespace, "8443")
 			oldHandlerCert := tests.EnsurePodsCertIsSynced(fmt.Sprintf("%s=%s", v1.AppLabel, "virt-handler"), flags.KubeVirtInstallNamespace, "8186")
 			Expect(err).ToNot(HaveOccurred())
@@ -195,7 +324,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			By("repeatedly starting VMIs until virt-api and virt-handler certificates are updated")
 			Eventually(func() (rotated bool) {
 				vmi := tests.NewRandomVMIWithEphemeralDisk(cd.ContainerDiskFor(cd.ContainerDiskAlpine))
-				vmi = tests.RunVMIAndExpectLaunch(vmi, 60)
+				vmi = tests.RunVMI(vmi, 60)
 				Expect(console.LoginToAlpine(vmi)).To(Succeed())
 				err = virtClient.VirtualMachineInstance(vmi.Namespace).Delete(vmi.Name, &metav1.DeleteOptions{})
 				Expect(err).ToNot(HaveOccurred())
@@ -243,7 +372,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			RestClient().
 			Post().
 			Resource("virtualmachineinstances").
-			Namespace(tests.NamespaceTestDefault).
+			Namespace(util.NamespaceTestDefault).
 			Body(vmi).
 			Do(context.Background()).Get()
 		Expect(err).ToNot(HaveOccurred(), "Should create VMI")
@@ -297,7 +426,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				Expect(len(pods.Items)).To(BeNumerically(">", 0), "no kubevirt pods found")
 
 				By("finding all schedulable nodes")
-				schedulableNodesList := tests.GetAllSchedulableNodes(virtClient)
+				schedulableNodesList := util.GetAllSchedulableNodes(virtClient)
 				schedulableNodes := map[string]*k8sv1.Node{}
 				for _, node := range schedulableNodesList.Items {
 					schedulableNodes[node.Name] = node.DeepCopy()
@@ -495,7 +624,8 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 	Describe("[rfe_id:3187][crit:medium][vendor:cnv-qe@redhat.com][level:component]Prometheus Endpoints", func() {
 		var preparedVMIs []*v1.VirtualMachineInstance
 		var pod *k8sv1.Pod
-		var metricsIPs []string
+		var handlerMetricIPs []string
+		var controllerMetricIPs []string
 
 		pinVMIOnNode := func(vmi *v1.VirtualMachineInstance, nodeName string) *v1.VirtualMachineInstance {
 			if vmi == nil {
@@ -585,6 +715,15 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 		tests.BeforeAll(func() {
 			tests.BeforeTestCleanup()
 
+			By("Finding the virt-controller prometheus endpoint")
+			virtControllerLeaderPodName := getLeader()
+			leaderPod, err := virtClient.CoreV1().Pods(flags.KubeVirtInstallNamespace).Get(context.Background(), virtControllerLeaderPodName, metav1.GetOptions{})
+			Expect(err).ToNot(HaveOccurred(), "Should find the virt-controller pod")
+
+			for _, ip := range leaderPod.Status.PodIPs {
+				controllerMetricIPs = append(controllerMetricIPs, ip.IP)
+			}
+
 			// The initial test for the metrics subsystem used only a single VM for the sake of simplicity.
 			// However, testing a single entity is a corner case (do we test handling sequences? potential clashes
 			// in maps? and so on).
@@ -595,11 +734,11 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			// any node is fine, we don't really care, as long as we run all VMIs on it.
 			prepareVMIForTests(nodeName)
 
-			By("Finding the prometheus endpoint")
+			By("Finding the virt-handler prometheus endpoint")
 			pod, err = kubecli.NewVirtHandlerClient(virtClient).Namespace(flags.KubeVirtInstallNamespace).ForNode(nodeName).Pod()
 			Expect(err).ToNot(HaveOccurred(), "Should find the virt-handler pod")
 			for _, ip := range pod.Status.PodIPs {
-				metricsIPs = append(metricsIPs, ip.IP)
+				handlerMetricIPs = append(handlerMetricIPs, ip.IP)
 			}
 		})
 
@@ -730,7 +869,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			if netutils.IsIPv6String(ip) {
 				Skip("Skip testing with IPv6 until https://github.com/kubevirt/kubevirt/issues/4145 is fixed")
@@ -770,7 +909,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			Expect(err).ToNot(HaveOccurred(), "Should throttle HTTP access without unexpected errors")
 		},
 			table.Entry("[test_id:4140] by using IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by using IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6226] by using IPv6", k8sv1.IPv6Protocol),
 		)
 
 		table.DescribeTable("should include the metrics for a running VM", func(family k8sv1.IPFamily) {
@@ -778,7 +917,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			By("Scraping the Prometheus endpoint")
 			Eventually(func() string {
@@ -788,7 +927,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			}, 30*time.Second, 2*time.Second).Should(ContainSubstring("kubevirt"))
 		},
 			table.Entry("[test_id:4141] by using IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by using IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6227] by using IPv6", k8sv1.IPv6Protocol),
 		)
 
 		table.DescribeTable("should include the storage metrics for a running VM", func(family k8sv1.IPFamily, metricSubstring, operator string) {
@@ -796,7 +935,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			metrics := collectMetrics(ip, metricSubstring)
 			By("Checking the collected metrics")
@@ -814,21 +953,21 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			}
 		},
 			table.Entry("[test_id:4142] storage flush requests metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_flush_requests_total", ">="),
-			table.Entry("storage flush requests metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_flush_requests_total", ">="),
+			table.Entry("[test_id:6228] storage flush requests metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_flush_requests_total", ">="),
 			table.Entry("[test_id:4142] time (ms) spent on cache flushing metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_flush_times_ms_total", ">="),
-			table.Entry("time (ms) spent on cache flushing metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_flush_times_ms_total", ">="),
+			table.Entry("[test_id:6229] time (ms) spent on cache flushing metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_flush_times_ms_total", ">="),
 			table.Entry("[test_id:4142] I/O read operations metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_iops_read_total", ">="),
-			table.Entry("I/O read operations metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_iops_read_total", ">="),
+			table.Entry("[test_id:6230] I/O read operations metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_iops_read_total", ">="),
 			table.Entry("[test_id:4142] I/O write operations metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_iops_write_total", ">="),
-			table.Entry("I/O write operations metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_iops_write_total", ">="),
+			table.Entry("[test_id:6231] I/O write operations metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_iops_write_total", ">="),
 			table.Entry("[test_id:4142] storage read operation time metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_read_times_ms_total", ">="),
-			table.Entry("storage read operation time metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_read_times_ms_total", ">="),
+			table.Entry("[test_id:6232] storage read operation time metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_read_times_ms_total", ">="),
 			table.Entry("[test_id:4142] storage read traffic in bytes metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_read_traffic_bytes_total", ">="),
-			table.Entry("storage read traffic in bytes metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_read_traffic_bytes_total", ">="),
+			table.Entry("[test_id:6233] storage read traffic in bytes metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_read_traffic_bytes_total", ">="),
 			table.Entry("[test_id:4142] storage write operation time metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_write_times_ms_total", ">="),
-			table.Entry("storage write operation time metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_write_times_ms_total", ">="),
+			table.Entry("[test_id:6234] storage write operation time metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_write_times_ms_total", ">="),
 			table.Entry("[test_id:4142] storage write traffic in bytes metric by using IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_storage_write_traffic_bytes_total", ">="),
-			table.Entry("storage write traffic in bytes metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_write_traffic_bytes_total", ">="),
+			table.Entry("[test_id:6235] storage write traffic in bytes metric by using IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_storage_write_traffic_bytes_total", ">="),
 		)
 
 		table.DescribeTable("should include metrics for a running VM", func(family k8sv1.IPFamily, metricSubstring, operator string) {
@@ -836,7 +975,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			metrics := collectMetrics(ip, metricSubstring)
 			By("Checking the collected metrics")
@@ -848,15 +987,15 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			}
 		},
 			table.Entry("[test_id:4143] network metrics by IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_network_", ">="),
-			table.Entry("network metrics by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_network_", ">="),
+			table.Entry("[test_id:6236] network metrics by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_network_", ">="),
 			table.Entry("[test_id:4144] memory metrics by IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_memory", ">="),
-			table.Entry("memory metrics by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_memory", ">="),
+			table.Entry("[test_id:6237] memory metrics by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_memory", ">="),
 			table.Entry("[test_id:4553] vcpu wait by IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_vcpu_wait", "=="),
-			table.Entry("vcpu wait by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_vcpu_wait", "=="),
+			table.Entry("[test_id:6238] vcpu wait by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_vcpu_wait", "=="),
 			table.Entry("[test_id:4554] vcpu seconds by IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_vcpu_seconds", ">="),
-			table.Entry("vcpu seconds by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_vcpu_seconds", ">="),
+			table.Entry("[test_id:6239] vcpu seconds by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_vcpu_seconds", ">="),
 			table.Entry("[test_id:4556] vmi unused memory by IPv4", k8sv1.IPv4Protocol, "kubevirt_vmi_memory_unused_bytes", ">="),
-			table.Entry("vmi unused memory by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_memory_unused_bytes", ">="),
+			table.Entry("[test_id:6240] vmi unused memory by IPv6", k8sv1.IPv6Protocol, "kubevirt_vmi_memory_unused_bytes", ">="),
 		)
 
 		table.DescribeTable("should include VMI infos for a running VM", func(family k8sv1.IPFamily) {
@@ -864,7 +1003,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			metrics := collectMetrics(ip, "kubevirt_vmi_")
 			By("Checking the collected metrics")
@@ -895,7 +1034,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			}
 		},
 			table.Entry("[test_id:4145] by IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6241] by IPv6", k8sv1.IPv6Protocol),
 		)
 
 		table.DescribeTable("should include VMI phase metrics for all running VMs", func(family k8sv1.IPFamily) {
@@ -903,7 +1042,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			metrics := collectMetrics(ip, "kubevirt_vmi_")
 			By("Checking the collected metrics")
@@ -916,7 +1055,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			}
 		},
 			table.Entry("[test_id:4146] by IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6242] by IPv6", k8sv1.IPv6Protocol),
 		)
 
 		table.DescribeTable("should include VMI eviction blocker status for all running VMs", func(family k8sv1.IPFamily) {
@@ -924,7 +1063,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(controllerMetricIPs, family)
 
 			metrics := collectMetrics(ip, "kubevirt_vmi_non_evictable")
 			By("Checking the collected metrics")
@@ -936,7 +1075,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			}
 		},
 			table.Entry("[test_id:4148] by IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6243] by IPv6", k8sv1.IPv6Protocol),
 		)
 
 		table.DescribeTable("should include kubernetes labels to VMI metrics", func(family k8sv1.IPFamily) {
@@ -944,7 +1083,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			// Every VMI is labeled with kubevirt.io/nodeName, so just creating a VMI should
 			// be enough to its metrics to contain a kubernetes label
@@ -960,7 +1099,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			Expect(containK8sLabel).To(Equal(true))
 		},
 			table.Entry("[test_id:4147] by IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6244] by IPv6", k8sv1.IPv6Protocol),
 		)
 
 		// explicit test fo swap metrics as test_id:4144 doesn't catch if they are missing
@@ -969,7 +1108,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				libnet.SkipWhenNotDualStackCluster(virtClient)
 			}
 
-			ip := getSupportedIP(metricsIPs, family)
+			ip := getSupportedIP(handlerMetricIPs, family)
 
 			metrics := collectMetrics(ip, "kubevirt_vmi_memory_swap_")
 			var in, out bool
@@ -989,7 +1128,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 			Expect(out).To(BeTrue())
 		},
 			table.Entry("[test_id:4555] by IPv4", k8sv1.IPv4Protocol),
-			table.Entry("by IPv6", k8sv1.IPv6Protocol),
+			table.Entry("[test_id:6245] by IPv6", k8sv1.IPv6Protocol),
 		)
 	})
 
@@ -1034,7 +1173,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				vmi := tests.NewRandomVMI()
 
 				By("Starting a new VirtualMachineInstance")
-				obj, err := virtClient.RestClient().Post().Resource("virtualmachineinstances").Namespace(tests.NamespaceTestDefault).Body(vmi).Do(context.Background()).Get()
+				obj, err := virtClient.RestClient().Post().Resource("virtualmachineinstances").Namespace(util.NamespaceTestDefault).Body(vmi).Do(context.Background()).Get()
 				Expect(err).To(BeNil())
 				tests.WaitForSuccessfulVMIStart(obj)
 			}, 150)
@@ -1133,12 +1272,14 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				}, 15*time.Second, 1*time.Second).Should(Equal(true))
 			})
 
-			It("[test_id:6246] label nodes with cpu model and cpu features", func() {
+			It("[test_id:6246] label nodes with cpu model, cpu features and host cpu model", func() {
 				for _, node := range nodesWithKVM {
 					Expect(err).ToNot(HaveOccurred())
 					cpuModelLabelPresent := false
 					cpuFeatureLabelPresent := false
 					hyperVLabelPresent := false
+					hostCpuModelPresent := false
+					hostCpuRequiredFeaturesPresent := false
 					for key := range node.Labels {
 						if strings.Contains(key, v1.CPUModelLabel) {
 							cpuModelLabelPresent = true
@@ -1149,18 +1290,29 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 						if strings.Contains(key, v1.HypervLabel) {
 							hyperVLabelPresent = true
 						}
+						if strings.Contains(key, v1.HostModelCPULabel) {
+							hostCpuModelPresent = true
+						}
+						if strings.Contains(key, v1.HostModelRequiredFeaturesLabel) {
+							hostCpuRequiredFeaturesPresent = true
+						}
 
-						if cpuModelLabelPresent && cpuFeatureLabelPresent && hyperVLabelPresent {
+						if cpuModelLabelPresent && cpuFeatureLabelPresent && hyperVLabelPresent && hostCpuModelPresent &&
+							hostCpuRequiredFeaturesPresent {
 							break
 						}
 					}
-					Expect(cpuModelLabelPresent).To(Equal(true), "node "+node.Name+" doesn't contain cpu label")
-					Expect(cpuFeatureLabelPresent).To(Equal(true), "node "+node.Name+" doesn't contain feature label")
-					Expect(hyperVLabelPresent).To(Equal(true), "node "+node.Name+" doesn't contain hyperV label")
+
+					errorMessageTemplate := "node " + node.Name + " does not contain %s label"
+					Expect(cpuModelLabelPresent).To(BeTrue(), fmt.Sprintf(errorMessageTemplate, "cpu"))
+					Expect(cpuFeatureLabelPresent).To(BeTrue(), fmt.Sprintf(errorMessageTemplate, "feature"))
+					Expect(hyperVLabelPresent).To(BeTrue(), fmt.Sprintf(errorMessageTemplate, "hyperV"))
+					Expect(hostCpuModelPresent).To(BeTrue(), fmt.Sprintf(errorMessageTemplate, "host cpu model"))
+					Expect(hostCpuRequiredFeaturesPresent).To(BeTrue(), fmt.Sprintf(errorMessageTemplate, "host cpu required featuers"))
 				}
 			})
 
-			It("should set default obsolete cpu models filter when obsolete-cpus-models is not set in kubevirt config", func() {
+			It("[test_id:6247] should set default obsolete cpu models filter when obsolete-cpus-models is not set in kubevirt config", func() {
 				node := nodesWithKVM[0]
 
 				for key := range node.Labels {
@@ -1173,27 +1325,37 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				}
 			})
 
-			It("should set default min cpu model filter when min-cpu is not set in kubevirt config", func() {
+			It("[test_id:6248] should set default min cpu model filter when min-cpu is not set in kubevirt config", func() {
 				node := nodesWithKVM[0]
 
 				for key := range node.Labels {
 					Expect(key).ToNot(Equal(v1.CPUFeatureLabel+"apic"), "Node can't contain label with apic feature (it is feature of default min cpu)")
 				}
 			})
+
+			It("[test_id:6995]should expose tsc frequency and tsc scalability", func() {
+				node := nodesWithKVM[0]
+				Expect(node.Labels).To(HaveKey("cpu-timer.node.kubevirt.io/tsc-frequency"))
+				Expect(node.Labels).To(HaveKey("cpu-timer.node.kubevirt.io/tsc-scalable"))
+				Expect(node.Labels["cpu-timer.node.kubevirt.io/tsc-scalable"]).To(Or(Equal("true"), Equal("false")))
+				val, err := strconv.ParseInt(node.Labels["cpu-timer.node.kubevirt.io/tsc-frequency"], 10, 64)
+				Expect(err).ToNot(HaveOccurred())
+				Expect(val).To(BeNumerically(">", 0))
+			})
 		})
 
 		Context("advanced labelling", func() {
 			var originalKubeVirt *v1.KubeVirt
 
 			BeforeEach(func() {
-				originalKubeVirt = tests.GetCurrentKv(virtClient)
+				originalKubeVirt = util.GetCurrentKv(virtClient)
 			})
 
 			AfterEach(func() {
 				tests.UpdateKubeVirtConfigValueAndWait(originalKubeVirt.Spec.Configuration)
 			})
 
-			It("should update node with new cpu model label set", func() {
+			It("[test_id:6249] should update node with new cpu model label set", func() {
 				obsoleteModel := ""
 				node := nodesWithKVM[0]
 
@@ -1223,7 +1385,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				Expect(found).To(Equal(false), "Node can't contain label "+v1.CPUModelLabel+obsoleteModel)
 			})
 
-			It("should update node with new cpu model vendor label", func() {
+			It("[test_id:6250] should update node with new cpu model vendor label", func() {
 				nodes, err := virtClient.CoreV1().Nodes().List(context.Background(), metav1.ListOptions{})
 				Expect(err).ToNot(HaveOccurred())
 				for _, node := range nodes.Items {
@@ -1237,7 +1399,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				Fail("No node contains label " + v1.CPUModelVendorLabel)
 			})
 
-			It("should update node with new cpu feature label set", func() {
+			It("[test_id:6251] should update node with new cpu feature label set", func() {
 				node := nodesWithKVM[0]
 
 				numberOfLabelsBeforeUpdate := len(node.Labels)
@@ -1253,7 +1415,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				Expect(numberOfLabelsBeforeUpdate).ToNot(Equal(len(node.Labels)), "Node should have different number of labels")
 			})
 
-			It("should remove all cpu model labels (all cpu model are in obsolete list)", func() {
+			It("[test_id:6252] should remove all cpu model labels (all cpu model are in obsolete list)", func() {
 				node := nodesWithKVM[0]
 
 				obsoleteModels := nodelabellerutil.DefaultObsoleteCPUModels
@@ -1290,7 +1452,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 
 			BeforeEach(func() {
 				tests.BeforeTestCleanup()
-				originalKubeVirt = tests.GetCurrentKv(virtClient)
+				originalKubeVirt = util.GetCurrentKv(virtClient)
 
 			})
 
@@ -1320,7 +1482,7 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 				Expect(err).ToNot(HaveOccurred())
 			})
 
-			It("should remove old labeller labels and annotations", func() {
+			It("[test_id:6253] should remove old labeller labels and annotations", func() {
 				originalNode, err := virtClient.CoreV1().Nodes().Get(context.Background(), nodesWithKVM[0].Name, metav1.GetOptions{})
 				Expect(err).ToNot(HaveOccurred())
 
@@ -1388,26 +1550,26 @@ var _ = Describe("[Serial][sig-compute]Infrastructure", func() {
 
 func getLeader() string {
 	virtClient, err := kubecli.GetKubevirtClient()
-	tests.PanicOnError(err)
+	util.PanicOnError(err)
 
 	controllerEndpoint, err := virtClient.CoreV1().Endpoints(flags.KubeVirtInstallNamespace).Get(context.Background(), leaderelectionconfig.DefaultEndpointName, metav1.GetOptions{})
-	tests.PanicOnError(err)
+	util.PanicOnError(err)
 
 	var record resourcelock.LeaderElectionRecord
 	if recordBytes, found := controllerEndpoint.Annotations[resourcelock.LeaderElectionRecordAnnotationKey]; found {
 		err := json.Unmarshal([]byte(recordBytes), &record)
-		tests.PanicOnError(err)
+		util.PanicOnError(err)
 	}
 	return record.HolderIdentity
 }
 
 func getNewLeaderPod(virtClient kubecli.KubevirtClient) *k8sv1.Pod {
 	labelSelector, err := labels.Parse(fmt.Sprint(v1.AppLabel + "=virt-controller"))
-	tests.PanicOnError(err)
+	util.PanicOnError(err)
 	fieldSelector := fields.ParseSelectorOrDie("status.phase=" + string(k8sv1.PodRunning))
 	controllerPods, err := virtClient.CoreV1().Pods(flags.KubeVirtInstallNamespace).List(context.Background(),
 		metav1.ListOptions{LabelSelector: labelSelector.String(), FieldSelector: fieldSelector.String()})
-	tests.PanicOnError(err)
+	util.PanicOnError(err)
 	leaderPodName := getLeader()
 	for _, pod := range controllerPods.Items {
 		if pod.Name != leaderPodName {
@@ -1501,3 +1663,40 @@ func getMetricKeyForVmiDisk(keys []string, vmiName string, diskName string) stri
 	}
 	return ""
 }
+
+func getDownwardMetrics(vmi *v1.VirtualMachineInstance) (*api.Metrics, error) {
+	res, err := console.SafeExpectBatchWithResponse(vmi, []expect.Batcher{
+		&expect.BSnd{S: `sudo vm-dump-metrics 2> /dev/null` + "\n"},
+		&expect.BExp{R: `(?s)(<metrics>.+</metrics>)`},
+	}, 5)
+	if err != nil {
+		return nil, err
+	}
+	metricsStr := res[0].Match[2]
+	metrics := &api.Metrics{}
+	Expect(xml.Unmarshal([]byte(metricsStr), metrics)).To(Succeed())
+	return metrics, nil
+}
+
+func getTimeFromMetrics(metrics *api.Metrics) int {
+
+	for _, m := range metrics.Metrics {
+		if m.Name == "Time" {
+			val, err := strconv.Atoi(m.Value)
+			Expect(err).ToNot(HaveOccurred())
+			return val
+		}
+	}
+	Fail("no Time in metrics XML")
+	return -1
+}
+
+func getHostnameFromMetrics(metrics *api.Metrics) string {
+	for _, m := range metrics.Metrics {
+		if m.Name == "HostName" {
+			return m.Value
+		}
+	}
+	Fail("no hostname in metrics XML")
+	return ""
+}
